import json
import os
import shlex
from typing import Any, Dict, List, Union

NotebookDict = Dict[str, Any]

# TODO: This file is a copy-pasta from https://github.com/valohai/jupyhai
# TODO: DRY between libs


def parse_ipynb(content_or_str: Union[str, NotebookDict]) -> NotebookDict:
    """
    "Smartly" parse content that contains a notebook.
    * If a string, it's first JSON deserialized.
    * If it's a "wrapped" dict (i.e. contains "type" == "notebook" and "content"), unwraps the content
    * Asserts the content smells like a notebook ("nbformat")
    :param content: See above.
    :return: Notebook data.
    """
    if isinstance(content_or_str, str):
        content = json.loads(content_or_str)
    else:
        content = content_or_str
    if not isinstance(content, dict):
        raise ValueError("Ipynb not a dict")
    if content.get("type") == "notebook":
        content = content["content"]

    nbformat = content.get("nbformat")
    if not isinstance(nbformat, int):
        raise ValueError(f"Nbformat value {nbformat} invalid")
    return dict(content)


def get_notebook_source_code(contents: NotebookDict) -> str:
    source = [
        cell["source"] for cell in contents["cells"] if cell["cell_type"] == "code"
    ]

    # Some notebook versions store it as list of rows already. Some as single string.
    source = [row if isinstance(row, list) else row.split("\n") for row in source]

    # Even when it was a list, the linefeeds are still there.
    source = [row.rstrip() for sublist in source for row in sublist]

    # Strip magics like "!pip install tensorflow"
    source = [row for row in source if not row.startswith("!")]

    return "\n".join(source)


def get_notebook_command(notebook_relative_path: str) -> List[str]:
    notebook_dir, notebook_name = os.path.split(notebook_relative_path)
    papermill_command = " ".join(
        [
            "papermill -k python3 -f /valohai/config/parameters.yaml",
            shlex.quote(
                f"/valohai/repository/{notebook_relative_path.replace(os.sep, '/')}"
            ),
            shlex.quote(f"/valohai/outputs/{notebook_name.replace(os.sep, '/')}"),
        ]
    )
    return ["pip install -r requirements.txt", papermill_command]



from valohai.internals.distributed_config.distributed_config import DistributedConfig
from valohai.internals.distributed_config.member import Member

__all__ = [
    "DistributedConfig",
    "Member",
]



from typing import Optional

from valohai.internals import global_state
from valohai.internals.global_state_loader import load_global_state
from valohai.types import InputDict, ParameterDict


def prepare(
    *,
    step: str,
    default_parameters: Optional[ParameterDict] = None,
    default_inputs: Optional[InputDict] = None,
    image: Optional[str] = None,
    environment: Optional[str] = None,
) -> None:
    """Define the name of the step and it's required inputs, parameters and Docker image

    Has dual purpose:
    - Provide default values for inputs, parameters and Docker image so the user code can be executed
    - Provide entry-point for the parser that generates/updates valohai.yaml integration file

    :param step: Step name for valohai.yaml
    :param default_parameters: Dict of parameters and default values
    :param default_inputs: Dict of inputs with (list of) default URIs
    :param image: Default docker image
    :param environment: Default environment ID or slug

    """

    global_state.step_name = step
    global_state.image_name = image
    global_state.environment = environment

    load_global_state(
        default_inputs_from_prepare=default_inputs,
        default_parameters_from_prepare=default_parameters,
    )



from importlib.abc import Loader
from typing import Callable, Optional

from papi import Papi
from valohai_yaml.objs import Config


def get_pipeline_from_source(source_path: str, old_config: Config) -> Config:
    """Gets the pipeline definition by executing the main() in the source Python file.

    The file is expected to contain:

    def main(config) -> Pipeline:
        pipeline = Pipeline(name="foo", config=config)
        # ... Pipeline definition ...
        return pipeline

    :param source_path: Path of the Python source code file containing the pipeline definition
    :param old_config: Currently config for the Valohai project (used for validation)

    """
    import importlib.util

    spec = importlib.util.spec_from_file_location(
        name="pipeline_source", location=source_path
    )
    if not spec:
        raise ValueError(f"Could not find import spec from {source_path}")
    module = importlib.util.module_from_spec(spec)
    loader: Optional[Loader] = spec.loader
    if not loader:
        raise ValueError("Spec has no loader")
    loader.exec_module(module)
    main: Optional[Callable[[Config], Papi]] = getattr(module, "main", None)
    if not main:
        raise AttributeError(f"{source_path} is missing main() method!")
    pipe = main(old_config)
    return Config(pipelines=[pipe.to_yaml()])



import json
from typing import TYPE_CHECKING, Any, Dict

from valohai import paths

if TYPE_CHECKING:
    import requests


def get_api_requests_kwargs(endpoint: str) -> Dict[str, Any]:
    """
    Get the "presigned call" dict for a given endpoint from the
    API JSON configuration file.  Will happily throw all sorts of
    exceptions e.g. if the API JSON file is missing or malformed.
    """
    with open(paths.get_api_config_path()) as json_file:
        api_config = json.load(json_file)
    value = api_config.get(endpoint)
    if not (isinstance(value, dict) and value.get("url")):
        raise ValueError(f"Invalid API config for {endpoint}")
    return value


def send_api_request(
    endpoint: str,
    **requests_kwargs: Any,
) -> "requests.Response":
    """
    Send an API request to the named endpoint.

    Will happily throw all sorts of exceptions; it is the caller's
    responsibility to handle them in a suitable way.

    :param endpoint: The endpoint to send the request to.  Will be
                     looked up in the API JSON configuration file.
    :param requests_kwargs: Any keyword arguments to pass to the
                            requests.request() function.  You can expect
                            `url` and `method` and likely `headers` to
                            have been set for you.
    """
    import requests

    requests_config = get_api_requests_kwargs(endpoint)
    requests_config.update(requests_kwargs)
    resp = requests.request(**requests_config)
    resp.raise_for_status()
    return resp



from typing import Any, Dict, List

from valohai.internals.distributed_config.member import Member
from valohai.internals.distributed_config.utils import rank_members


class DistributedConfig:
    def __init__(
        self,
        *,
        group_name: str,
        member_id: str,
        required_count: int,
        members: List[Member],
    ):
        self.group_name = group_name
        self.member_id = member_id
        self.required_count = required_count
        self.members = members

    @classmethod
    def from_json_data(cls, json_data: Dict[str, Any]) -> "DistributedConfig":
        members = [Member.from_json_data(m) for m in json_data["members"]]
        rank_members(members)
        return cls(
            group_name=json_data["config"]["group_name"],
            member_id=json_data["config"]["member_id"],
            required_count=json_data["config"]["required_count"],
            members=members,
        )



import argparse
import json
import os
import sys
from typing import Any, Dict, List, Optional, Set, Tuple, Union

from valohai_yaml.utils import listify

from valohai.internals import global_state
from valohai.internals.input_info import InputInfo
from valohai.internals.utils import string_to_bool
from valohai.paths import get_inputs_config_path, get_parameters_config_path
from valohai.types import InputDict, ParameterDict


def load_global_state(
    default_inputs_from_prepare: Optional[InputDict] = None,
    default_parameters_from_prepare: Optional[ParameterDict] = None,
) -> None:
    """Loads inputs & parameters and stores their value in the global_state

    Inputs & parameters (with values) can be defined in three places:
    - Call to valohai.prepare()
    - File-based config in inputs.json and parameters.json
    - Overridable values from CLI args

    It is possible that something is defined in all the three places simultaneously.
    Thus, we have a priority order for the final value:

    1. CLI override
    2. File-based config
    3. valohai.prepare()

    :param default_inputs_from_prepare: Dict of inputs and their values from valohai.prepare()
    :param default_parameters_from_prepare: Dict of parameters and their values from valohai.prepare()
    """

    inputs = {}
    parameters = {}

    # Load defaults from call to valohai.prepare()
    if default_inputs_from_prepare:
        inputs = sift_defaults(default_inputs_from_prepare)
    if default_parameters_from_prepare:
        parameters = sift_defaults(default_parameters_from_prepare)

    # Use inputs.json and parameters.json instead if found
    inputs = load_inputs_from_config() or inputs
    parameters = load_parameters_from_config() or parameters

    # Parse and override input & parameter values from CLI
    cli_inputs, cli_parameters = parse_overrides_from_cli(
        input_names=set(inputs),
        parameters=parameters,
    )
    inputs = {key: cli_inputs.get(key, value) for key, value in inputs.items()}
    final_parameters = {
        key: cli_parameters.get(key, value) for key, value in parameters.items()
    }

    # Inputs in various formats are converted into InputInfo(s)
    final_inputs = {key: convert_to_input_info(value) for key, value in inputs.items()}

    # Save results to global_state
    global_state.inputs = final_inputs
    global_state.parameters = final_parameters

    # Store an explicit flag to be used in load_global_state_if_necessary()
    global_state.loaded = True


def load_global_state_if_necessary() -> None:
    """Loads the global state if it is not already loaded

    It is possible that user hasn't called valohai.prepare() at all,
    but still tries to access inputs and parameters.

    This method is for making sure that the global_state is loaded before accessing them.

    """
    if not global_state.loaded:
        load_global_state()


def parse_overrides_from_cli(
    *,
    input_names: Set[str],
    parameters: ParameterDict,
) -> Tuple[Dict[str, List[str]], Dict[str, Any]]:
    """Override inputs and parameters from the command-line

    :param input_names: List of input names
    :param parameters: Dict of parameters and their values
    """

    parser = argparse.ArgumentParser()
    for name in input_names:
        parser.add_argument(f"--{name}", type=str, nargs="+")
    for name, value in parameters.items():
        # TODO: this does not properly handle `value` possibly being a default dict

        if isinstance(value, bool):
            # We need to fiddle booleans in a bit different way, since they are treated as flags per default
            parser.add_argument(f"--{name}", type=string_to_bool, nargs="?", const=True)
        else:
            parser.add_argument(f"--{name}", type=type(value))
    known_args, unknown_args = parser.parse_known_args()

    for unknown in unknown_args:
        print(  # noqa
            f"Warning: Unexpected command-line argument {unknown} found.",
            file=sys.stderr,
        )

    cli_inputs = sift_cli_inputs(known_args, input_names)
    cli_parameters = sift_cli_parameters(known_args, set(parameters.keys()))

    return cli_inputs, cli_parameters


def load_inputs_from_config() -> InputDict:
    """Load input values from the config file if it exists."""
    inputs = {}
    config_path = get_inputs_config_path()
    if os.path.isfile(config_path):
        with open(config_path) as json_file:
            inputs = json.load(json_file)
    return inputs


def load_parameters_from_config() -> ParameterDict:
    parameters = {}
    config_path = get_parameters_config_path()
    if os.path.isfile(config_path):
        with open(config_path) as json_file:
            parameters = json.load(json_file)
    return parameters


def sift_cli_inputs(
    args: argparse.Namespace, expected_keys: Set[str]
) -> Dict[str, List[str]]:
    """Sift inputs from all the command-line args

    :param expected_keys: List of expected input names
    """

    result = {}
    for name, values in vars(args).items():
        # Filter out any inputs that we weren't expecting
        if name not in expected_keys:
            continue

        values = [v for v in listify(values) if v is not None]
        if len(values) > 0:
            result[name] = values
    return result


def sift_cli_parameters(
    args: argparse.Namespace, expected_keys: Set[str]
) -> Dict[str, Any]:
    """Sift parameters from all the command-line args

    :param expected_keys: List of expected parameter names
    """

    return {
        name: value
        for name, value in vars(args).items()
        if name in expected_keys and value is not None
    }


def sift_defaults(values: Dict[str, Any]) -> Dict[str, Any]:
    """Returns the default values which user defined in .prepare()

    Works for both inputs and parameters.

    We sift from two alternatives (dict and nested dict):

    valohai.prepare(
        default_parameters={"my-param", 123},
        default_inputs={"my-input", "https://lol.png"}
        )
    valohai.prepare(
        default_parameters={"my-param", {"default": 123}},
        default_inputs={"my-input", {"default": "https://lol.png"}}
        )

    :param values: Default values from the .prepare() method
    """

    result = {}
    for key, value in values.items():
        if isinstance(value, dict):
            if "default" in value:
                result[key] = value["default"]
            else:
                raise ValueError(f"No default value defined for {key}")
        else:
            result[key] = value
    return result


def convert_to_input_info(input: Union[str, List[str], Dict[str, Any]]) -> InputInfo:
    """Converts inputs from different formats into an InputInfo

    Inputs can be defined in either .prepare() or inputs.json

    Inputs defined in .prepare() are a List[str] or str
    Inputs defined in inputs.json are a Dict

    Ultimately we want have all of them converted as InputInfo(s)

    :param inputs: Single input as List[str] or Dict
    """
    if isinstance(input, (list, str)):
        return InputInfo.from_urls_and_paths(input)
    return InputInfo.from_json_data(input)



__version__ = "0.3.1"

import papi

from valohai.controller_api import set_status_detail
from valohai.inputs import inputs
from valohai.internals.global_state import distributed
from valohai.metadata import logger
from valohai.outputs import outputs
from valohai.parameters import parameters
from valohai.prepare_impl import prepare

Pipeline = papi.Papi

__all__ = [
    "distributed",
    "inputs",
    "logger",
    "outputs",
    "parameters",
    "prepare",
    "Pipeline",
    "set_status_detail",
]



import glob
import os
from typing import Any, Dict, Iterable, List, Optional, Union

from valohai_yaml.utils import listify

from valohai.internals.download import download_url
from valohai.internals.download_type import DownloadType
from valohai.internals.utils import uri_to_filename
from valohai.paths import get_inputs_path


class FileInfo:
    def __init__(
        self,
        *,
        name: str,
        uri: Optional[str] = None,
        path: Optional[str] = None,
        size: Optional[int] = None,
        checksums: Optional[Dict[str, str]] = None,
        metadata: Optional[List[Dict[str, Any]]] = None,
        datum_id: Optional[str] = None,
    ) -> None:
        self.name = str(name)
        self.uri = str(uri) if uri else None
        self.checksums = dict(checksums) if checksums else {}
        self.path = str(path) if path else None
        self.size = int(size) if size else None
        self.metadata = list(metadata) if metadata else []
        self.datum_id = str(datum_id) if datum_id else None

    def is_downloaded(self) -> Optional[bool]:
        return bool(self.path and os.path.isfile(self.path))

    def download(self, path: str, force_download: bool = False) -> None:
        if not self.uri:
            raise ValueError("Can not download file with no URI")
        self.path = download_url(
            self.uri, os.path.join(path, self.name), force_download
        )
        # TODO: Store size & checksums if they become useful

    @classmethod
    def from_json_data(cls, json_data: Dict[str, Any]) -> "FileInfo":
        return cls(
            name=json_data["name"],
            uri=json_data.get("uri"),
            path=json_data.get("path"),
            size=json_data.get("size"),
            checksums=json_data.get("checksums"),
            metadata=json_data.get("metadata"),
            datum_id=json_data.get("datum_id"),
        )


class InputInfo:
    def __init__(self, files: Iterable[FileInfo]):
        self.files = list(files)

    def is_downloaded(self) -> bool:
        if not self.files:
            return False

        return all(f.is_downloaded() for f in self.files)

    def download_if_necessary(
        self, name: str, download: DownloadType = DownloadType.OPTIONAL
    ) -> None:
        if (
            download == DownloadType.ALWAYS
            or not self.is_downloaded()
            and download == DownloadType.OPTIONAL
        ):
            path = get_inputs_path(name)
            os.makedirs(path, exist_ok=True)
            for f in self.files:
                f.download(path, force_download=(download == DownloadType.ALWAYS))

    @classmethod
    def from_json_data(cls, json_data: Dict[str, Any]) -> "InputInfo":
        return cls(
            files=[FileInfo.from_json_data(d) for d in json_data.get("files", ())]
        )

    @classmethod
    def from_urls_and_paths(cls, urls_and_paths: Union[str, List[str]]) -> "InputInfo":
        files = []

        for value in listify(urls_and_paths):
            # TODO: this cast shouldn't be required,
            #       but listify()'s types are wonky with new mypy
            value = str(value)
            if "://" not in value:  # The string is a local path
                for path in glob.glob(value):
                    files.append(
                        FileInfo(
                            name=os.path.basename(path),
                            uri=None,
                            path=path,
                            size=None,
                            checksums=None,
                        )
                    )
            else:  # The string is an URL
                files.append(
                    FileInfo(
                        name=uri_to_filename(value),
                        uri=value,
                        path=None,
                        size=None,
                        checksums=None,
                    )
                )

        return cls(files=files)



import warnings


def set_status_detail(status_detail: str) -> None:
    """
    Attempt to set the status detail of the current execution.
    """
    try:
        from valohai.internals.api_calls import send_api_request

        send_api_request(
            endpoint="set_status_detail",
            json={"status_detail": status_detail},
        )
    except FileNotFoundError:  # No API configuration file, no need to warn
        pass
    except Exception as exc:
        warnings.warn(f"Status detail send failed: {exc}")



import json
import warnings
from typing import List, Optional

from valohai import paths
from valohai.internals.distributed_config import DistributedConfig, Member


class Distributed:
    """
    Distributed toolkit accessed through `valohai.distributed`.
    """

    _config: Optional[DistributedConfig] = None

    def is_distributed_task(self) -> bool:
        # not a property to mimic `is_running_in_valohai`
        try:
            return bool(self.config.group_name)
        except (FileNotFoundError, json.JSONDecodeError):
            return False
        except Exception as exc:
            warnings.warn(f"Failed to parse distributed config: {exc}")
            return False

    @property
    def group_name(self) -> str:
        return self.config.group_name

    @property
    def member_id(self) -> str:
        return self.config.member_id

    @property
    def rank(self) -> Optional[int]:
        return self.me().rank

    @property
    def required_count(self) -> int:
        return self.config.required_count

    def members(self) -> List[Member]:
        return self.config.members

    def member(self, member_id: str) -> Member:
        for member in self.members():
            if member.member_id == member_id:
                return member
        raise RuntimeError(f"No member with id {member_id}")

    def me(self) -> Member:
        return self.member(member_id=self.config.member_id)

    def master(self) -> Member:
        for member in self.members():
            if member.is_master:
                return member
        raise RuntimeError("No master member found")

    @property
    def config(self) -> DistributedConfig:
        if not self._config:
            with open(self._get_config_path()) as json_file:
                json_data = json.load(json_file)
            self._config = DistributedConfig.from_json_data(json_data)
        return self._config

    def _get_config_path(self) -> str:
        return paths.get_distributed_config_path()

    def flush_state(self) -> None:
        self._config = None



from typing import Optional

from valohai.internals import global_state, vfs
from valohai.internals.download_type import DownloadType
from valohai.internals.global_state_loader import load_global_state_if_necessary
from valohai.internals.input_info import InputInfo


def get_input_vfs(
    name: str,
    process_archives: bool = True,
    download_type: DownloadType = DownloadType.OPTIONAL,
) -> vfs.VFS:
    v = vfs.VFS()
    ii = get_input_info(name)
    if ii:
        ii.download_if_necessary(name, download_type)
        for file_info in ii.files:
            assert file_info.path
            vfs.add_disk_file(
                v,
                name=file_info.name,
                path=file_info.path,
                process_archives=process_archives,
            )

    return v


def get_input_info(name: str) -> Optional[InputInfo]:
    load_global_state_if_necessary()
    return global_state.inputs.get(name, None)



def uri_to_filename(uri: str) -> str:
    return uri.rpartition("/")[-1].split("?", 1)[0]


def string_to_bool(value: str) -> bool:
    if value.lower() == "false":
        return False
    return True



from typing import TYPE_CHECKING, Dict, Iterable, Union

if TYPE_CHECKING:
    from valohai.internals.distributed_config import Member


def rank_members(members: Iterable["Member"]) -> None:
    """Add ranks to members in-place."""
    mapping = compute_member_id_ranks([m.member_id for m in members])
    for m in members:
        m.rank = mapping[m.member_id]


def compute_member_id_ranks(member_ids: Iterable[str]) -> Dict[str, int]:
    """Given member ids, return member id to rank mapping."""
    id_to_sortable: Dict[str, Union[str, int]]
    try:
        id_to_sortable = {member_id: int(member_id) for member_id in member_ids}
    except ValueError:
        # if we fail to parse all member ids as an integer, just sort as string
        id_to_sortable = {member_id: member_id for member_id in member_ids}
    return {
        member_id: index
        for index, (member_id, sort_value) in enumerate(
            sorted(id_to_sortable.items(), key=lambda kv: kv[1])
        )
    }



from valohai.internals.parameters import get_parameter_value, supported_types


class Parameter:
    def __init__(self, name: str, default: supported_types = None) -> None:
        self.name = str(name)
        self.default = default

    @property
    def value(self) -> supported_types:
        return get_parameter_value(self.name, self.default)


parameters = Parameter
